      subroutine psqf2(qfunc,qfcoef,rho,lll,
     +  rinner,mesh,r,rab,a,b,kkbeta,ifprt,nang,
     +  nqf,qtryc,ibfix,rfix,qfix,nfix,ifqopt,
     +  phi1,phi2,ib,jb,idim1,idim3,idim6,idim8)
c
c----------------------------------------------------------------------------
c
      implicit double precision(a-h,o-z)
c
c     written from psqf1 by dominic king-smith in november 1992
c     modified by kurt stokbro
c     modified further by d vanderbilt 8/97
c
c     arrays like 'targ' that are dimensioned idm8 have the
c     following structure:
c       1 to nqf             taylor functions 1, r**2, r**4 etc
c       nqf+1 to nqf+ncon-1  normal constraints (see below)
c       nqf+ncon             optional extra constraint set by nfix
c     where ncon=4 and ncon=6 for ifqopt=2 and 3 respectively:
c       ifqopt=2:  3 normal constraints are (value,d1,norm)
c       ifqopt=3:  5 normal constraints are (value,d1,d2,d3,norm)
c     where d1,d2,d3 are 1st, 2nd, 3rd derivative
c
c     ------------------------------------------------------
c     notes on qfunc and qfcoef:
c     ------------------------------------------------------
c     since q_ij(r) is the product of two orbitals like
c     psi_{l1,m1}^star * psi_{l2,m2}, it can be decomposed by
c     total angular momentum l, where l runs over | l1-l2 | ,
c     | l1-l2 | +2 , ... , l1+l2.  (l=0 is the only component
c     needed by the atomic program, which assumes spherical
c     charge symmetry.)
c     
c     recall  qfunc(r) = y1(r) * y2(r)  where y1 and y2 are the
c     radial parts of the wave functions defined according to
c     
c       psi(r-vec) = (1/r) * y(r) * y_lm(r-hat)  .
c     
c     for each total angular momentum l, we pseudize qfunc(r)
c     inside rc as:
c     
c       qfunc(r) = r**(l+2) * [ a_1 + a_2*r**2 + a_3*r**4 + ...]
c     
c     in such a way as to match qfunc and its
c             1'st derivative at rc   (ifqopt=2)
c             1'st-3'rd derivs at rc  (ifqopt=3)
c     and to preserve
c     
c       int_{0}^{r_c} dr r**l * qfunc(r)   ,
c     
c     i.e., to preserve the l'th moment of the charge.  also,
c     depending on nfix,ibfix,etc, there may be one more constraint
c     to obtain a particular value at a particular radius (as a trick
c     for avoiding negative charge densities).  the rest of
c     the coeffs are determined by minimizing 
c
c       integral (from qtryc to infinity) dq q**2 * qfunc(q)**2
c     where
c       qfunc(q)=int_{0}^{r_c} dr qfunc(r) * j_l(qr)
c
c     where qtryc and the number of coeffs nqf are input.
c     
c     qfunc has been set inside rc to correspond to this pseudized
c     version using the minimal l, namely l = | l1-l2 | (e.g., l=0
c     for diagonal elements).  the coefficients a_i (i=1,2,3)
c     are stored in the array qfcoef(i,l+1,j,k) for each l so that
c     the correctly pseudized versions of qfunc can be reconstructed
c     for each l.  (note that for given l1 and l2, only the values
c     l = | l1-l2 | , | l1-l2 | +2 , ... , l1+l2 are ever used.)
c
c     ------------------------------------------------------
c
c.....in order to avoid passing many unnecessary arrays, use following
      parameter(idm6=5,nqfx=10,nconx=6,ngx=101,idm8=nqfx+nconx)
c
c.....logarithmic radial mesh information
      dimension r(idim1),rab(idim1)
c.....q functions and q pseudization coefficients
      dimension qfunc(idim1),qfcoef(idim8,idim6)
c.....q functions and q pseudization cutoffs
      dimension rinner(idim6),icut(idm6)
c.....q functions and l-dependent pseudization arrays for plot
      dimension qfplot(1000,idm6)
c.....extra fixed points to avoid negative densities when pseudizing q_ij
      dimension ibfix(idim3),rfix(idim3),qfix(idim3)
c.....the phi functions
      dimension phi1(idim1),phi2(idim1)
c.....angular momentum information
      dimension lll(idim3)
c.....scratch
      dimension rho(idim1)
      dimension targ(idm8),bb(idm8,idm8,idm6),targinv(idm8)
      dimension qn(ngx,nqfx,idm6),qlf(ngx),g(ngx)
      logical   tinit,tcon
c
      integer stderr
      common /files/ stderr,input,iout,ioae,iplot,iologd,iops
c
c     data to be saved to avoid reinitialization
c
      save qn,bb,g
      data tinit/.false./
c
c-----------------------------------------------------------------------
c
c                      i n i t i a l i z a t i o n
c
      if (ifprt .gt. 1 .or. (.not. tinit)) write (iout,*) 
     + ' entering subroutine psqf2, tinit = ',tinit
c
c     set up qfplot for original unpseudized qfs
      if (ib .eq. jb) then
        nline = 1
        do 90 i=1,idim1
          qfplot(i,nline)=qfunc(i)
 90     continue
      endif
c
c     compute maximum and minimum angular momenta
c
      lmin = iabs ( lll(ib) - lll(jb) )
      lmax = iabs ( lll(ib) + lll(jb) )
c
      write(iout,'(100a,/,6(a,i2),/,100a)') ('=',i=1,100),
     +  ' ib =',ib,' , lll(ib) =',lll(ib),' , jb = ',jb,
     +  ' , lll(jb) =',lll(jb),' , lmin =',lmin,' , lmax =',lmax,
     +  ('-',i=1,100)
c

c     =========================================================
c     set variables that control array index idm8
c     =========================================================
c
      if (ifqopt.eq.2) ncon=4
      if (ifqopt.eq.3) ncon=6
c
      nnorm=nqf+ncon-1
      nxtra=nqf+ncon
c
c     ================
c     ckeck dimensions
c     ================
c
      if ( nqf .le. ncon-1 ) then
        write(iout,*) '***error in subroutine psqf2'
        write(iout,*) 'nqf must be .gt.',ncon-1,' but nqf=',nqf
        call exit(1)
      endif
c
      if(nqfx.lt.nqf) then
        write(iout,*) '***error: in subroutine psqf2, nqfx too small'
        write(iout,*) 'nqfx, nqf = ',nqfx,nqf
        call exit(1)
      endif
c
      if(nconx.lt.ncon) then
        write(iout,*) '***error: in subroutine psqf2, nconx too small'
        write(iout,*) 'nconx, ncon = ',nconx,ncon
        call exit(1)
      endif
c
      if(idm6.lt.2*nang-1) then
        write(iout,*) '***error: in subroutine psqf2, idm6 too small'
        write(iout,*) 'idm6, 2*nang-1 = ',idm6,2*nang-1
        call exit(1)
      endif
c
c     set the value of pi
c
      pi=4.0d0*atan(1.0d0)
      tpi=2.0*pi
      tpi3=tpi**3
      fpi=4.0*pi
c
c     =========================
c     possible extra constraint
c     =========================
c
      tcon = .false.
      if ( ib .eq. jb ) then
c
        do 10 ifix = 1,nfix
c
          if ( ib .eq. ibfix(ifix) ) then
c
            write(iout,*) 'extra constraint q_ij pseudization',
     +        ' constraint for ibeta =',ib
            write(iout,'(a,f12.7,a,f12.7)') 
     +        'will ensure q_ij passes through',qfix(ifix),
     +        ' at r =',rfix(ifix)
c
            ifcur = ifix
c
            tcon = .true.
            if ( nqf .le. ncon ) then
              write(iout,*) '***error subroutine psqf2'
              write(iout,*) 'nqf .le.',ncon,' will produce singular bb'
              call exit(1)
            endif
c
c           set up the last row and column of the bb matrix
c           corresponding to the extra constraint
c
            do 20 iqf = 1,nqf
c
              iqft = 2*(iqf-1)
c
              bb(nxtra,iqf,1) = rfix(ifix) ** iqft
c
   20       continue
c
            do 30 iqf = nqf+1,nxtra
              bb(nxtra,iqf,1) = 0.0d0
   30       continue
c
            do 40 iqf = 1,nxtra
              bb(iqf,nxtra,1) = bb(nxtra,iqf,1)
   40       continue
c
            targ(nxtra) = qfix(ifix) / rfix(ifix) ** 2
c
          endif
c
   10   continue
c
      endif
c
c     =============================
c     one time initialization tasks
c     =============================
c
      if (.not. tinit) then
c
c     ==========================================
c     prepare the uniformly spaced g-vector list
c     ==========================================
c
      dg=qtryc/dfloat(ngx-1)
      do 200 iq=1,ngx
        g(iq)=dble(iq-1)*dg
  200 continue
c
c     ====================
c     compute the qn array
c     ====================
c
      write(iout,*)'nqf,nqfx,idm6',nqf,nqfx,idm6
      call setqn(qn,g,rinner,nqf,2*nang-1,ngx,nqfx,idm6)
      write(iout,*)'exit setqn psqf2'
c
      do 250 ltp=1,2*nang-1
c
c     =========================
c     find interpolation points
c     =========================
c
      do 100 i2 = kkbeta,1,-1
        if ( r(i2) .lt. rinner(ltp) ) goto 110
  100 continue
c
  110 icut(ltp) = i2
c
c     =====================
c     construct matrices bb
c     =====================
c
        do 260 iqf=1,nnorm
        do 260 jqf=1,nnorm
          bb(iqf,jqf,ltp)=0.0d0
  260   continue
c
c       ========================================================
c       set parts of bb arising from imposing normal constraints
c       ========================================================
c
        do 270 iqf=1,nqf
c
          iqft=2*(iqf-1)
c
c         (1) value at rinner is preserved
c
          bb(nqf+1,iqf,ltp)=rinner(ltp)**iqft
c
c         (2) derivative at rinner is preserved
c
          bb(nqf+2,iqf,ltp)=iqft*rinner(ltp)**(iqft-1)
c
          if (ifqopt.eq.3) then
c
c           (3) 2nd derivative at rinner is preserved
c
            bb(nqf+3,iqf,ltp)=(iqft*(iqft-1))*rinner(ltp)**(iqft-2)
c
c           (4) 3rd derivative at rinner is preserved
c
            bb(nqf+4,iqf,ltp)=(iqft*(iqft-1)*(iqft-2))*
     +           rinner(ltp)**(iqft-3)
c
          endif

c         (5) integral of r**(2*l+2)*rho should be preserved
c
          lmtp=2*(ltp-1+iqf)+1
          bb(nnorm,iqf,ltp)=rinner(ltp)**lmtp/dfloat(lmtp)
c
c         symmetrical part
          do jqf=nqf+1,nnorm
            bb(iqf,jqf,ltp)=bb(jqf,iqf,ltp)
          end do
c
  270   continue
c
c       set parts of bb to do with smoothing of qfunc
c
        do 280 iqf=1,nqf
        do 280 jqf=1,nqf
          do 290 iq=1,ngx
            rho(iq)=g(iq)*g(iq)*qn(iq,iqf,ltp)*qn(iq,jqf,ltp)
  290     continue
          call simpagn(ngx,g,rho,asum,0)
          lijtm=2*(ltp-1+iqf+jqf)-1
          bb(iqf,jqf,ltp)=tpi3*rinner(ltp)**lijtm/dfloat(lijtm)-asum
  280   continue
c
  250 continue
c
c     end one time initialization events
c
      tinit = .true.
      endif
c
c-----------------------------------------------------------------------
c
c     b e g i n  l o o p  o v e r  l t o t
c
      do 300 ltot = 0,2*(nang-1)
c
        ltp = ltot + 1
c
c       only need to pseudize for lmin,lmin+2 ...lmax-2,lmax
c
        if ( ltot .lt. lmin .or. ltot .gt. lmax .or. 
     +    mod(ltot-lmin,2) .eq. 1 ) then
c
          do 310 iqf=1,nqf
            qfcoef(iqf,ltp)=0.0d0
  310     continue
c
        else
c
c         =============
c         construct qlf 
c         =============
c
c         note that 
c
c           qlf(g) = 4 * pi * int_{rinner}^{r_c} qfunc(r) j_l(gr) dr
c
          do 320 iq=1,ngx
            do 330 i=1,mesh
              rho(i)=qfunc(i)*bessel(g(iq)*r(i),ltot)
  330       continue
            call simp(rho,r,rab,a,b,rinner(ltp),kkbeta,asum2,idim1)
            call radin(mesh,rho,rab,asum1,idim1)
            qlf(iq)=fpi*(asum1-asum2)
  320     continue
c
c         ============
c         compute tqlf
c         ============
c
c         note that
c
c           tqlf = int_{0}^{qtryc} g**2 qlf(g)**2 dg
c
          do 340 iq=1,ngx
            rho(iq)=(g(iq)*qlf(iq))**2
  340     continue
          call simpagn(ngx,g,rho,tqlf,0)
c
c         ===========
c         compute tqr
c         ===========
c
c         note that
c
c           tqr = 8 * pi**3 int_{rinner}^{r_c} qfunc(r)**2 / r**2 dr
c
          rho(1)=0.0
          do 350 i=2,mesh
            rho(i)=(qfunc(i)/r(i))**2
  350     continue
          call simp(rho,r,rab,a,b,rinner(ltp),kkbeta,asum2,idim1)
          call radin(mesh,rho,rab,asum1,idim1)
          tqr=tpi3*(asum1-asum2)
c
c         ================================================
c         construct the vector targ which depends on qfunc
c         ================================================
c
          do 400 i = 2,kkbeta
            rho(i) = qfunc(i) / r(i)**(ltot+2)
  400     continue
c
          if (ifqopt.eq.2) then
c
            i2 = icut(ltp)
            i1 = i2 - 1
            i3 = i2 + 1
            i4 = i2 + 2
c
            call polyn(rinner(ltp),val,r(i1),rho(i1),-1)
            call polyn(rinner(ltp),val,r(i1),rho(i1), 0)
            call polyn(rinner(ltp), d1,r(i1),rho(i1),+1)
c
            targ(nqf+1) = val
            targ(nqf+2) = d1
c
c           possibly save val and d1 for later print out
            if ( ltot .eq. lmin ) then
              valmin = val
              d1min = d1
            endif
c
          elseif (ifqopt.eq.3) then
c
            i2 = icut(ltp)
            i1 = i2 - 5
            i3 = i2 + 1
            i4 = i2 + 6
c
            call polyn2(rinner(ltp),val,r(i1),rho(i1),-1)
            call polyn2(rinner(ltp),val,r(i1),rho(i1), 0)
            targ(nqf+1)=val
            do i = 1,ncon-3
               call polyn2(rinner(ltp), d1,r(i1),rho(i1),i)
               targ(nqf+1+i) = d1
            enddo
c
c           possibly save val and d1 for later print out
            if ( ltot .eq. lmin ) then
              valmin = val
              d1min = targ(nqf+2)
              d2min = targ(nqf+3)
              d3min = targ(nqf+4)
              d1min = d1
            endif
c
          else
            call exit(1)
          endif
c
          rho(1) = 0.d0
          do 410 i = 2,kkbeta
            rho(i) = qfunc(i) * r(i) ** ltot
  410     continue
c
          call simp(rho,r,rab,a,b,rinner(ltp),kkbeta,conae,idim1)
c
          targ(nnorm)=conae
c
          do 420 iqf=1,nqf
            do 430 iq=1,ngx
              rho(iq)=g(iq)*g(iq)*qn(iq,iqf,ltp)*qlf(iq)
  430       continue
            call simpagn(ngx,g,rho,asum,0)
            targ(iqf)=asum
  420     continue
c
c         ====================
c         invert bb for qfcoef
c         ====================
c
          if ( tcon .and. ltot .eq. 0 ) then
            ncons = ncon
          else
            ncons = ncon-1
          endif
c
          call simeq(idm8,nqf+ncons,bb(1,1,ltp),targ,targinv)
          do 435 iqf=1,nqf
            qfcoef(iqf,ltp)=targinv(iqf)
  435     continue
c
c         =================================
c         possiple print out of information
c         =================================
c
          if (ifprt.gt.1) then
            write(iout,'(//,a,i3,a,2i3,//)') 'psqf2: results ltot =',
     +        ltot,' with ibeta,jbeta =',ib,jb
            write(iout,*) ' the bb matrix = '
            do 440 i=1,nqf
              write(iout,490) (bb(i,j,ltp),j=1,nqf)
  440       continue
            write (iout,*) ' the targ vector is '
            write (iout,490) (targ(i),i=1,nqf)
            write (iout,*) ' qfcoef for ltp = ',ltp
            write (iout,490) (qfcoef(i,ltp),i=1,nqf)
            write (iout,*) ' the lagrange mutipliers = '
            write (iout,490) (targinv(i),i=nqf+1,nqf+ncons)
  490       format (8(1pd13.5))
c
          endif
c
c         ==================================
c         calculate and print residual chisq
c         ==================================
c
c         note: chisq = fpi**2 int_{qtryc}^{inf} g**2 qfunc(g)**2 dg
c
          chisq=tqr-tqlf
          do 450 i=1,nqf
            chisq=chisq-2.0*targ(i)*targinv(i)
            do 460 j=1,nqf
              chisq=chisq+targinv(i)*bb(i,j,ltp)*targinv(j)
  460       continue
  450     continue
c
c         =====================================
c         compute percentage weight above qtryc
c         =====================================
c
c         note: total = fpi**2 int_{0}^{inf}  g**2 qfunc(g)**2 dg
c               total = tpi**3 int_{0}^{r_c}  qfunc(r)**2 / r**2 dr
c
c         load rho for r <= r(i2) with qfunc
          call setqf(qfcoef,rho,r,nqf,ltot,i2,idim1,idim6,idim8)
c
          do 600 ir = i3,mesh
            rho(ir) = qfunc(ir)
  600     continue
c
          rho(1) = 0.0d0
          do 610 ir = 2,mesh
            rho(ir) = ( rho(ir) / r(ir) ) ** 2
  610     continue
          call radin(mesh,rho,rab,total,idim1)
          total = total * tpi**3
c
c         ==================================================
c         check constraint int_{0}^{r_c} r**ltot qfunc(r) dr
c         ==================================================
c
c         load rho with pseudized form for r <= r(i2)
          call setqf(qfcoef,rho,r,nqf,ltot,i2,idim1,idim6,idim8)
c
          rho(1)=0.d0
          do 630 ir = 2,i2
            rho(ir) = rho(ir) * r(ir) ** ltot
  630     continue
c
          do 620 ir = i2,kkbeta
            rho(ir) = qfunc(ir) * r(ir) ** ltot
  620     continue
c
          call simp(rho,r,rab,a,b,rinner(ltp),kkbeta,conps,idim1)
c
c         summary of results
          write(iout,'(a,8x,a,13x,a,12x,a,11x,a,13x,a)') 
     +      'ltot','chisq','total','residual','conae','conps'
          write(iout,'(i3,3x,5(g15.7,3x))') 
     +       ltot,chisq,total,chisq/total,conae,conps
c
c--------------------------------------------------------------
c     set up qfplot for l-dependent pseudized q functions
c
          if (ib .eq. jb) then
            nline = nline + 1
            call setqf(qfcoef,rho,r,nqf,ltot,i2,
     +               idim1,idim6,idim8)
            do 91 i=1,i2
              qfplot(i,nline)=rho(i)
 91         continue
            write(iout,*) 'psqf2',nline,qfplot(1,nline)
            do 92 i=i3,idim1
              qfplot(i,nline)=qfunc(i)
 92         continue
          endif
c
        endif
c
c     end loop over tot ang momentum components
  300 continue
c
c       print the real-space qfunctions and Fourier transforms
c
      if (ib .eq. jb) then
        call prqf1(ifprt,qfplot,kkbeta,nline,r,a,b,idim1,idim3)
c
        write(stderr,*) 'fourier transforming the qfunctions'
c
c       fourier transform and print reciprocal-space q functions
        call fanalq1(r,rab,qfplot,nline,mesh,ifprt,
     +    idim1,idim3)
c
      endif
c
c-----------------------------------------------------------------------
c
c     s e t  q f u n c  f o r  l t o t  =  l m i n
c
      lminp = lmin +1
      i2 = icut(lminp)

      if (ifqopt.eq.2) then
c
        i1 = i2 - 1
        i3 = i2 + 1
        i4 = i2 + 2
        call setqf(qfcoef,qfunc,r,nqf,lmin,i2,idim1,idim6,idim8)
c
c       run some checks
        do ir = i1,i4
          rho(ir) = qfunc(ir) / r(ir)**(lmin+2)
        end do
c
        lminp = lmin +1
        call polyn(rinner(lminp),val,r(i1),rho(i1),-1)
        call polyn(rinner(lminp),val,r(i1),rho(i1), 0)
        call polyn(rinner(lminp), d1,r(i1),rho(i1),+1)
c
        write(iout,'(a,2f9.5)') 'target and achieved values of qfunc',
     +    valmin,val
        write(iout,'(a,2f9.5)') 'target and achieved values of dqfunc',
     +    d1min,d1
c
      elseif (ifqopt.eq.3) then
c
        i1 = i2 - 5
        i3 = i2 + 1
        i4 = i2 + 6
        call setqf(qfcoef,qfunc,r,nqf,lmin,i2,idim1,idim6,idim8)
c
c       run some checks
        do ir = i1,i4
          rho(ir) = qfunc(ir) / r(ir)**(lmin+2)
        end do
c
        lminp = lmin +1
        call polyn2(rinner(lminp),val,r(i1),rho(i1),-1)
        call polyn2(rinner(lminp),val,r(i1),rho(i1), 0)
        call polyn2(rinner(lminp), d1,r(i1),rho(i1),+1)
        call polyn2(rinner(lminp), d2,r(i1),rho(i1),+2)
        call polyn2(rinner(lminp), d3,r(i1),rho(i1),+3)
c
        write(iout,*) 'for lp=',lminp
        write(iout,'(a,2f11.5)')
     +    'target and achieved values of qfunc',valmin,val
        write(iout,'(a,2f11.5)')
     +    'target and achieved values of dqfunc',d1min,d1
        write(iout,'(a,2f11.5)')
     +    'target and achieved values of d2qfunc',d2min,d2
        write(iout,'(a,2f11.5)')
     +    'target and achieved values of d3qfunc',d3min,d3
c
      else
        call exit(1)
      endif
c
c     possible test of the constraint agreement
      if ( tcon ) then
c
        do 560 j3 = 3,i2
          if ( rfix(ifcur) .lt. r(j3) ) goto 570
  560   continue
c
  570   continue
c
        call polyn(rfix(ifcur),val,r(j3-2),qfunc(j3-2),-1)
        call polyn(rfix(ifcur),val,r(j3-2),qfunc(j3-2), 0)
c
        write(iout,'(a,4f9.5)') 'values of r about rfix    ',
     +    (r(j),j=j3-2,j3+1)
        write(iout,'(a,4f9.5)') 'values of q_ij about qfix ',
     +    (qfunc(j),j=j3-2,j3+1)
        write(iout,'(2a,2f12.7)') 'extra constraint:',
     +    ' target and achieved',qfix(ifcur),val
c
      endif
c
      if (ifprt.gt.0) then
        write (iout,*) ' i1, i2, i3, i4 =',i1,i2,i3,i4
        write (iout,120) ' r1, r2, r3, r4 =',r(i1),r(i2),r(i3),r(i4)
        write (iout,120) 'qf1,qf2,qf3,qf4 =',qfunc(i1),qfunc(i2),
     +       qfunc(i3),qfunc(i4)
  120   format(a18,4f8.4)
      endif
c
      return
      end
